Attribute VB_Name = "NeoMutations"
Option Explicit

Private Type MutationsType
  PointPerUnit As Long 'per kilocycle
  PointCycle As Long 'the cycle we are goin to point mutate during
  PointMean As Single 'average length of DNA to change
  PointStdDev As Single 'std dev of this length

  ReproduceTotalPerUnit As Long
  Reproduce As Long 'a 1 in X chance to mutate when we reproduce

  ReversalPerUnit As Long
  ReversalLengthMean As Single
  ReversalLengthStdDev As Single

  CopyErrorPerUnit As Long
  InsertionPerUnit As Long
  AmplificationPerUnit As Long
  MajorDeletionPerUnit As Long
  MinorDeletionPerUnit As Long
End Type

'1-(perbot+1)^(1/DNALength) = per unit
'1-(1-perunit)^DNALength = perbot

Public Const PointUP As Integer = 0 'expressed as 1 chance in X per kilocycle per bp
Public Const CopyErrorUP As Integer = 6
Public Const MinorDeletionUP As Integer = 1
Public Const InsertionUP As Integer = 3
Public Const AmplificationUP As Integer = 4
Public Const MajorDeletionUP As Integer = 5
Public Const ReversalUP As Integer = 2
Public Const DeltaUP As Integer = 7

'NEVER allow anything after end, which must be = DNALen
'ALWAYS assume that DNA is sized right
'ALWAYS size DNA correctly when mutating

Private Function EraseUnit(ByRef unit As block)
  unit.tipo = 0
  unit.value = 0
End Function

Public Function MakeSpace(ByRef DNA() As block, beginning As Integer, length As Integer, Optional DNALength As Integer = -1) As Boolean
  'add length elements after beginning.  Beginning doesn't move places
  'returns true if the space was created,
  'false otherwise

  Dim t As Integer
  If DNALength < 0 Then DNALength = DnaLen(DNA)
  If length < 1 Or beginning < 1 Or beginning > DNALength - 1 Then
    MakeSpace = False
    Exit Function
  End If
  
  MakeSpace = True

  ReDim Preserve DNA(DNALength + length)

  For t = DNALength To beginning + 1 Step -1
    DNA(t + length) = DNA(t)
    EraseUnit DNA(t)
  Next t
End Function

Public Sub Delete(ByRef DNA() As block, beginning As Integer, elements As Integer, Optional DNALength As Integer = -1)
  'delete elements starting at beginning
  Dim t As Integer
  If DNALength < 0 Then DNALength = DnaLen(DNA)
  If elements < 1 Or beginning < 1 Or beginning > DNALength - 1 Then Exit Sub
  If elements + beginning > DNALength - 1 Then elements = DNALength - 1 - beginning

  For t = beginning + elements To DNALength
    DNA(t - elements) = DNA(t)
  Next t

  DNALength = DnaLen(DNA)
  ReDim Preserve DNA(DNALength)
End Sub

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Public Sub Mutate(robn As Integer)
  With rob(robn)
    'If .Mutables.mutarray(PointUP) > 0 Then PointMutation robn
    'If .Mutables.mutarray(DeltaUP) > 0 Then DeltaMut robn
  End With
End Sub

'
'Private Sub PointMutation(ByVal robn As Long)
'  'assume the bot has a positive (>0) mutarray value for this
'  With rob(robn)
'    If .age = 0 Then PointMutWhereAndWhen Rnd, robn, .PointMutBP
'
'    While .age = .PointMutCycle
'      ChangeDNA rob(robn).DNA, .PointMutBP, _
'        Gauss(.MutStdDev(0, PointUP), .MutMean(0, PointUP))
'      PointMutWhereAndWhen Rnd, robn, .PointMutBP
'    Wend
'  End With
'End Sub
'
'Private Sub PointMutWhereAndWhen(randval As Single, robn As Long, Optional offset As Long = 0)
'  Dim result As Long
'  If randval = 0 Then randval = 0.0001
'  With rob(robn)
'  result = offset + Fix(Log(randval) / Log(1 - 1 / .mutarray(PointUP)))
'  .PointMutBP = result Mod .DnaLen
'  .PointMutCycle = .age + result / .DnaLen
'  End With
'End Sub
'
'Private Sub DeltaMut(robn As Integer)
'  With rob(robn)
'
'  If Rnd > 1 - 1 / .mutarray(DeltaUP) Then
'    If Rnd > .MutMean(1, DeltaUP) Then
'      'mutate just one
'      Dim temp As Integer
'
'      temp = Rnd * 20
'
'      '!!!!!!!!!!!!!!!!!
'      .mutarray(temp) = Gauss(.mutarray(temp))
'    Else
'      'mutate all
'    End If
'  End If
'  End With
'End Sub
'
'Private Sub DeltaMutSpecific(robn As Integer, place As Integer)
'  rob(robn).mutarray(place) = Gauss(rob(robn).mutarray(place), something)
'End Sub
'
'Private Sub ChangeDNA(ByRef DNA() As block, nth As Integer, Optional length As Long = 1)
'  'currently only mutates the value.  Later we may change it to accept changes
'  'in type
'
'  With rob(robn)
'
'  Dim t As Long
'  For t = nth To nth + length
'    If t >= .DnaLen Then Exit Sub 'don't mutate end either
'    If .DNA(t).tipo = 10 Then Exit Sub 'mutations can't cross control barriers
'    Dim old As Long
'
'    'If Rnd() < .MutMean(1, PointUP) Then
'      'mutate value
'      old = .DNA(t).value
'
'      If .DNA(t).tipo = 0 Or .DNA(t).tipo = 1 Then '(number or *number)
'        Do
'          .DNA(t).value = Gauss(.DNA(t).value / 10, .DNA(t).value)
'        Loop While .DNA(t).value = old
'
'        .LastMutDetail = "Changed " + TipoDetok(.DNA(t).tipo) + " from " + _
'          old + " to " + .DNA(t).value + " at position " + t + " during cycle " + _
'          SimOpts.TotRunCycle + endl + .LastMutDetail
'      Else
'        Dim max As Long
'        Dim temp As String
'        Dim bp As block
'
'        'find max legit value
'        'this should really be done a better way
'        bp.tipo = .DNA(t).tipo
'        Do
'          max = max + 1
'          bp.value = max
'          Parse temp, bp
'        While temp <> ""
'
'        Do
'          .DNA(t).value = Rnd * (max - 1) + 1
'        Loop While .DNA(t).value = old
'
'        Dim Name As String
'        Dim oldname As String
'        Dim bp As block
'
'        bp.tipo = .DNA(t).tipo
'        bp.value = old
'
'        Parse Name, .DNA(t)
'        Parse oldname, bp
'
'        .LastMutDetail = "Changed " + TipoDetok(.DNA(t).tipo) + " from " + _
'          oldname + " to " + Name + " at position " + t + " during cycle " + _
'          SimOpts.TotRunCycle + endl + .LastMutDetail
'      End If
'    'Else
'    '  'mutate type
'    '  'valid types are 0 through 9
'    '
'    '
'    'End If
'  Next t
'  End With
'End Sub

Public Function delgene(n As Integer, g As Integer) As Boolean
  Dim k As Integer, t As Integer
  k = CountGenes(rob(n).DNA)
  If g > 0 And g <= k Then
    DeleteSpecificGene rob(n).DNA, g
    delgene = True
    'makecondlist n
    makeoccurrlist n
    countconditions n
  End If
End Function

Public Sub DeleteSpecificGene(ByRef DNA() As block, k As Integer)
  Dim i As Integer, f As Integer
  
  i = GenePos(DNA, k)
  If i < 0 Then Exit Sub
  f = NextStop(DNA, i)
  Delete DNA, i, f - i
End Sub
